<?php
// Own namespace
namespace Org\Mxchange\CoreFramework\Filesystem\Index;

// Import framework stuff
use Org\Mxchange\CoreFramework\Bootstrap\FrameworkBootstrap;
use Org\Mxchange\CoreFramework\Filesystem\File\BaseBinaryFile;
use Org\Mxchange\CoreFramework\Filesystem\Index\IndexableFile;
use Org\Mxchange\CoreFramework\Generic\UnsupportedOperationException;
use Org\Mxchange\CoreFramework\Index\Indexable;
use Org\Mxchange\CoreFramework\Utils\Crypto\CryptoUtils;

// Import SPL stuff
use \BadMethodCallException;
use \InvalidArgumentException;
use \SplFileInfo;

/**
 * An index file class
 *
 * @author		Roland Haeder <webmaster@ship-simu.org>
 * @version		0.0.0
 * @copyright	Copyright (c) 2007, 2008 Roland Haeder, 2009 - 2021 Core Developer Team
 * @license		GNU GPL 3.0 or any newer version
 * @link		http://www.ship-simu.org
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
class IndexFile extends BaseBinaryFile implements IndexableFile {
	/**
	 * Protected constructor
	 *
	 * @return	void
	 */
	private function __construct () {
		// Call parent constructor
		parent::__construct(__CLASS__);
	}

	/**
	 * Creates an instance of this File class and prepares it for usage
	 *
	 * @param	$fileInfoInstance	An instance of a SplFileInfo class
	 * @param	$indexInstance	An instance of a Indexable class
	 * @return	$indexFileInstance	An instance of an IndexableFile class
	 */
	public final static function createIndexFile (SplFileInfo $fileInfoInstance, Indexable $indexInstance) {
		// Get a new instance
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('INDEX-FILE: fileInfoInstance[%s]=%s,indexInstance=%s - CALLED!', get_class($fileInfoInstance), $fileInfoInstance, $indexInstance->__toString()));
		$indexFileInstance = new IndexFile();

		// Set file instance here for callbacks
		$indexFileInstance->setIndexInstance($indexInstance);

		// Expand file name with .idx
		$indexInfoInstance = new SplFileInfo(sprintf('%s.idx', $fileInfoInstance->__toString()));

		// Init this abstract file
		$indexFileInstance->initFile($indexInfoInstance);

		// Init counters and gaps array
		$indexFileInstance->initCountersGapsArray();

		// Return the prepared instance
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('INDEX-FILE: indexFileInstance=%s - EXIT!', $indexFileInstance->__toString()));
		return $indexFileInstance;
	}

	/**
	 * Flushes the file header
	 *
	 * @return	void
	 * @throws	BadMethodCallException	If this->indexInstance is not properly set
	 */
	public function flushFileHeader () {
		// Validate call
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('INDEX-FILE: CALLED!');
		if (!($this->getIndexInstance() instanceof Indexable)) {
			// Index instance not set
			throw new BadMethodCallException('this->indexInstance[] is not properly set.');
		}

		// Call block instance
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('INDEX-FILE: Calling this->indexInstance->flushFileHeader() ...');
		$this->getIndexInstance()->flushFileHeader();

		// Trace message
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('INDEX-FILE: EXIT!');
	}

	/**
	 * Pre-allocates file (if enabled) with some space for later faster write access.
	 *
	 * @param	$type	Type of the file
	 * @return	void
	 * @throws	InvalidArgumentException	If a parameter is empty
	 * @throws	BadMethodCallException	If this->indexInstance is not properly set
	 */
	public function preAllocateFile (string $type) {
		// Is it enabled?
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('INDEX-FILE: type=%s - CALLED!', $type));
		if (empty($type)) {
			// Empty type
			throw new InvalidArgumentException('Parameter "type" is empty');
		} elseif (!($this->getIndexInstance() instanceof Indexable)) {
			// Index instance not set
			throw new BadMethodCallException('this->indexInstance[] is not properly set.');
		}

		// Message to user
		self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('INDEX-FILE: Pre-allocating file ...');

		// Calculate minimum block length
		$minimumBlockLength = $this->getIndexInstance()->calculateMinimumBlockLength();

		// Call protected method
		$this->preAllocateFileByTypeLength($type, $minimumBlockLength);

		// Trace message
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('INDEX-FILE: EXIT!');
	}

	/**
	 * Checks wether the current entry is valid (not at the end of the file).
	 * This method will return true if an emptied (nulled) entry has been found.
	 *
	 * @return	$isValid	Whether the next entry is valid
	 * @throws	UnexpectedValueException	If some value is not expected
	 * @throws	BadMethodCallException	If this->indexInstance is not properly set
	 */
	public function isValid () {
		// Validate call
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('INDEX-FILE: CALLED!');
		if (!($this->getIndexInstance() instanceof Indexable)) {
			// Index instance not set
			throw new BadMethodCallException('this->indexInstance[] is not properly set.');
		}

		// Get length from index
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('INDEX-FILE: Calling this->indexInstance->calculateMinimumBlockLength() ...');
		$length = $this->getIndexInstance()->calculateMinimumBlockLength();

		// Call protected method
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('INDEX-FILE: Calling this->isValidByLength(%d) ...', $length));
		$isValid = $this->isValidByLength($length);

		// Return result
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('INDEX-FILE: isValid=%d - EXIT!', intval($isValid)));
		return $isValid;
	}

	/**
	 * Reads next "block" of bytes into $currentBlock field. THis method loads
	 * the whole file into memory when the file is just freshly initialized
	 * (only zeros in it).
	 *
	 * @return	void
	 */
	protected function readNextBlock () {
		// First calculate minimum block length
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('INDEX-FILE: this->seekPosition=%d - CALLED!', $this->getSeekPosition()));
		$length = $this->getIndexInstance()->calculateMinimumBlockLength();

		// Call protected method
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('INDEX-FILE: Calling parent::readNextBlockByLength(%d) ...', $length));
		parent::readNextBlockByLength($length);

		// Trace message
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('INDEX-FILE: EXIT!');
	}

	/**
	 * Reads the file header
	 *
	 * @return	void
	 */
	public function readFileHeader () {
		// Call index class' method
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('INDEX-FILE: Calling this->indexInstance->readIndexHeader() - CALLED!');
		$this->getIndexInstance()->readIndexHeader();

		// Trace message
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('INDEX-FILE: EXIT!');
	}

	/**
	 * Writes given value to the file and returns a hash and gap position for it
	 *
	 * @param	$stackName	Group identifier
	 * @param	$value		Value to be added to the stack
	 * @return	$data		Hash and gap position
	 * @throws	InvalidArgumentException	If a parameter is not valid
	 * @throws	BadMethodCallException	If this->indexInstance is not properly set
	 */
	public function writeValueToFile (string $stackName, $value) {
		// Validate parameter
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('INDEX-FILE: stackName=%s,value[]=%s - CALLED!', $stackName, gettype($value)));
		if (empty($stackName)) {
			// Throw IAE
			throw new InvalidArgumentException('Parameter "stackName" is empty');
		} elseif (is_object($value) || is_resource($value)) {
			// Not wanted here
			throw new InvalidArgumentException(sprintf('value[]=%s is not stackable in files', gettype($value)));
		} elseif (!($this->getIndexInstance() instanceof Indexable)) {
			// Index instance not set
			throw new BadMethodCallException('this->indexInstance[] is not properly set.');
		}

		// Encode/convert the value into a "binary format"
		$encoded = StringUtils::encodeData($value);

		// Get a strong hash for the "encoded" data
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('INDEX-FILE: encoded=%s', $encoded));
		$hash = CryptoUtils::hash($encoded);

		// Then write it to the next free gap
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('INDEX-FILE: hash=%s', $hash));
		$data = $this->getIndexInstance()->writeDataToFreeGap($stackName, $hash, $encoded);

		// Return info
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('INDEX-FILE: data[]=%s - EXIT!', gettype($data)));
		return $data;
	}

	/**
	 * Writes given raw data to the file and returns a gap position and length
	 *
	 * @param	$stackName	Group identifier
	 * @param	$hash		Hash from encoded value
	 * @param	$encoded	Encoded value to be written to the file
	 * @return	$data		Gap position and length of the raw data
	 * @throws	UnsupportedOperationException	If this method is called
	 */
	public function writeDataToFreeGap (string $stackName, string $hash, string $encoded) {
		self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('INDEX-FILE: stackName=' . $stackName . ',hash=' . $hash . ',encoded()=' . strlen($encoded));
		throw new UnsupportedOperationException(array($this, __FUNCTION__), self::EXCEPTION_UNSPPORTED_OPERATION);
	}

}
